#![feature(let_chains)]
use std::{collections::{HashSet, HashMap}, thread::current, borrow::Borrow};

#[derive(Clone, Hash, Eq, PartialEq)]
pub struct Point {
    pub x: usize,
    pub y: usize
}

pub struct HeightMap {
    map: Vec<Vec<char>>,
    best_steps: HashMap<Point, usize>,
    start: Point,
    end: Point
}

impl HeightMap {
    pub fn get_neighbors(& self, point: &Point) -> Vec<Point> {
        let mut neighbors = Vec::new();
        if point.x > 0 {
            neighbors.push(Point { x: point.x - 1, y: point.y});
        }
        if point.x + 1< self.map[0].len() {
            neighbors.push(Point { x: point.x + 1, y: point.y})
        }
        if point.y > 0 {
            neighbors.push(Point { y: point.y - 1, x: point.x})
        }
        if point.y + 1 < self.map.len() {
            neighbors.push(Point { y: point.y + 1, x: point.x})
        }

        return neighbors
    }

    pub fn acceptable_next(curr: &char) -> char {
        if *curr == 'S' {
            return 'a'
        }
        if *curr == 'z' {
            return 'E'
        }

        return ((*curr as u8) + 1).try_into().unwrap()
    }

    pub fn best_possible_dist(&self, curr: &Point) -> usize {
        self.end.x.abs_diff(curr.x) + self.end.y.abs_diff(curr.y)
    }
}


// if !(neighbor_val == HeightMap::acceptable_next(&curr_val) || (neighbor_val <= curr_val) && (neighbor_val != 'E' && neighbor_val != 'S')) {
//     continue;

fn neighbor_has_acceptbale_value(our_value: char, neighbor_value: char, acceptable_value: char) -> bool {
    if neighbor_value == acceptable_value {
        return true
    }
    if neighbor_value != 'S' && neighbor_value != 'E' && neighbor_value <= our_value {
        return true
    } 
    false
}

fn dijkstra(map: HeightMap, unvisited: &mut HashSet<Point>, distances: &mut HashMap<Point, usize>, end: &Point) {
    while unvisited.contains(end) {
        let curr_point = unvisited.iter().reduce(|curr, new| if distances[curr] < distances[new] { curr} else {new}).unwrap().clone();
        // let (_curr_point, _curr_dist) = distances.iter().reduce(|old, new| if new.1 < old.1 { new } else { old }).unwrap();
        let curr_dist = distances[&curr_point];
        // println!("checking {}, {}", curr_point.x, curr_point.y);

        if curr_point == *end {
            return;
        }

        let our_value = map.map[curr_point.y][curr_point.x];
        let allowed_neighbor_value = HeightMap::acceptable_next(&our_value);
        let neighbors: Vec<Point> = map.get_neighbors(&curr_point).into_iter().filter(
            |neighbor| neighbor_has_acceptbale_value(our_value, map.map[neighbor.y][neighbor.x], allowed_neighbor_value)
        ).collect();

        for neighbor in neighbors {
            if distances[&neighbor] > curr_dist + 1 {
                distances.insert(neighbor, curr_dist + 1);
            }
        }

        unvisited.remove(&curr_point);
    }
}

pub fn part_one(input: &str) -> Option<usize> {
    let _map: Vec<Vec<char>> = input.lines().map(|line| line.chars().collect()).collect();
    let mut end_coord = Point { x: 0, y: 0 };
    let mut start_coord = Point { x: 0, y: 0 };
    let mut to_visit: HashSet<Point> = HashSet::new();
    let mut distances: HashMap<Point, usize> = HashMap::new();


    for (r_idx, row) in _map.iter().enumerate() {
        for (col_idx, c) in row.iter().enumerate() {
            to_visit.insert(Point {x: col_idx, y: r_idx});
            distances.insert(Point {x: col_idx, y: r_idx}, usize::MAX);
            if *c == 'E' {
                end_coord = Point {y: r_idx, x: col_idx};
            } else if *c == 'S' {
                start_coord = Point {y: r_idx, x: col_idx};
            }
        }
    }

    distances.insert(start_coord.clone(), 0);

    let mut map = HeightMap {
        best_steps: HashMap::new(),
        map: _map,
        start: start_coord.clone(),
        end: end_coord.clone()
    };
    dijkstra(map, &mut to_visit, &mut distances, &end_coord);
    return Some(distances[&end_coord]);

}

pub fn part_two(input: &str) -> Option<u32> {
    None
}

fn main() {
    let input = &advent_of_code::read_file("inputs", 12);
    advent_of_code::solve!(1, part_one, input);
    advent_of_code::solve!(2, part_two, input);
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_part_one() {
        let input = advent_of_code::read_file("examples", 12);
        assert_eq!(part_one(&input), Some(31));
    }

    #[test]
    fn test_part_two() {
        let input = advent_of_code::read_file("examples", 12);
        assert_eq!(part_two(&input), None);
    }
}
